/***************************************************************************
 *   Copyright (C) 2004 by Tyler Montbriand                                *
 *   tsm@accesscomm.ca                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <stdio.h>
#include <stdlib.h>

#include "SDL_audioin.h"

static volatile int fd=-1;
static volatile SDL_audiostatus ainstat=SDL_AUDIO_STOPPED;

// Converts SDL audio types to UNIX ones
extern int mask_in[],mask_out[],samplesize[];

static SDL_AudioSpec localspec;
static SDL_Thread *ainthread=NULL;

int ainthread_func(void *);

static int          DSP_GetFrameSize(int fd);
static int          DSP_SetFrameSize(int fd, int framesize);
static int          DSP_SetChannels(int fd, int channels);
static int          DSP_SetRate(int fd, int rate);
static unsigned int DSP_GetMask(int fd);
static int          DSP_SetFormat(int fd, unsigned int format);

int SDL_OpenAudioIn(SDL_AudioSpec *requested, SDL_AudioSpec *result)
{
  const char *dev="/dev/dsp";
  int mask=0,n;

  if(ainstat!=SDL_AUDIO_STOPPED)
  {
    fprintf(stderr,"Warning, audio input already initialized\n");
    return(-1);
  }

  if(requested==NULL)
    return(-1);

  localspec=(*requested);
  requested=&localspec;

  fd=open(dev,O_RDONLY);
  if(fd<0)
  {
    fprintf(stderr,"Couldn't open /dev/dsp\n");
    return(-1);
  }

  mask=DSP_GetMask(fd);
  if(mask==0)
  {
    fprintf(stderr,"No formats supported?  Not good.\n");
    close(fd);
    return(-1);
  }

  for(n=0; mask_in[n]>=0; n++)
    if((mask_in[n]==requested->format)&&(mask&mask_out[n]))
    {
      requested->size=requested->samples*requested->channels*samplesize[n];
      requested->silence=0;

      if((DSP_SetFormat(fd,mask_out[n])<0)||
        (DSP_SetChannels(fd,requested->channels)<0)||
        (DSP_SetFrameSize(fd,requested->size)<0)||
        (DSP_SetRate(fd,requested->freq)<0))
      {
        fprintf(stderr,"Couldn't properly initialize %s\n",dev);
	close(fd);
	return(-1);
      }

      ainstat=SDL_AUDIO_PAUSED;
      ainthread=SDL_CreateThread(ainthread_func,NULL);
      if(ainthread==NULL)
      {
        fprintf(stderr,"Error, couldn't create audio thread\n");
	ainstat=SDL_AUDIO_STOPPED;
	close(fd);
	return(-1);
      }

      if(result!=NULL) (*result)=(*requested);
      // Success!
      return(0);
    }

  fprintf(stderr,"Format not supported\n");
  close(fd);
  fd=-1;
  return(-1);
}

void SDL_PauseAudioIn(int pause)
{
  if(ainstat!=SDL_AUDIO_STOPPED)
  {
    if(pause)
      ainstat=SDL_AUDIO_PAUSED;
    else
      ainstat=SDL_AUDIO_PLAYING;
  }
}

void SDL_CloseAudioIn()
{
  int status;
  if(ainstat!=SDL_AUDIO_STOPPED)
  {
    ainstat=SDL_AUDIO_STOPPED;
    SDL_WaitThread(ainthread,&status);
    if(status!=0)
      fprintf(stderr,"Warning, audio thread exited with status %d\n",status);
    ainthread=NULL;
  }
}

int ainthread_func(void *dat)
{
  Uint8 *buf=NULL;

  if(ainstat==SDL_AUDIO_STOPPED)
    return(-1);

  buf=(Uint8 *)malloc(localspec.size);
  if(buf==NULL)
  {
    fprintf(stderr,"Error!  Couldn't allocate buffer\n");
    ainstat=SDL_AUDIO_STOPPED;
    return(-1);
  }

  while(ainstat!=SDL_AUDIO_STOPPED)
  {
    ssize_t bytes=0;
    bytes=read(fd,buf,localspec.size);

    if(bytes<0)
    {
      ainstat=SDL_AUDIO_STOPPED;
      fprintf(stderr,"Error!  Sound device died\n");
      free(buf);
      return(-1);
    }

    if(ainstat==SDL_AUDIO_PLAYING)
    {
      if(localspec.callback!=NULL)
        localspec.callback(localspec.userdata,buf,localspec.size);
    }
  }

  free(buf);
  return(0);
}

unsigned int DSP_GetMask(int fd)
{
  unsigned int mask=0;
  ioctl(fd,SNDCTL_DSP_GETFMTS,&mask);
  return(mask);
}

int DSP_SetChannels(int fd, int channels)
{
  return(ioctl(fd,SNDCTL_DSP_CHANNELS,&channels));
}

int DSP_SetRate(int fd, int rate)
{
  return(ioctl(fd,SNDCTL_DSP_SPEED,&rate));
}

int DSP_GetFrameSize(int fd)
{
  int framesize=0;
  ioctl(fd,SNDCTL_DSP_GETBLKSIZE,&framesize);
  return(framesize);
}

int DSP_SetFrameSize(int fd, int framesize)
{
  return(ioctl(fd,SNDCTL_DSP_SETFRAGMENT,&framesize));
}

int DSP_SetFormat(int fd, unsigned int format)
{
  return(ioctl(fd,SNDCTL_DSP_SETFMT,&format));
}

int mask_in[]=
{
  AUDIO_U8
, AUDIO_S8
, AUDIO_U16LSB
, AUDIO_S16LSB
, AUDIO_U16MSB
, AUDIO_S16MSB
, -1
};
int mask_out[]=
{
   AFMT_U8
,  AFMT_S8
,  AFMT_U16_LE
,  AFMT_S16_LE
,  AFMT_U16_BE
,  AFMT_S16_BE
, -1
};
int samplesize[]=
{ 1,1,2,2,2,2,-1 };
